#!/usr/bin/perl -w
#
# Nagios configuration generator (WLCG probe based)
# Copyright (c) 2007 Emir Imamagic
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

use strict;
use Getopt::Long;
use Socket;
use XML::Parser;
use LWP::UserAgent;
use Net::LDAP;
use Sys::Hostname;
use Date::Format;
use Env qw(MYPROXY_SERVER SITE_NAME);

use wlcgprobes qw($WLCG_NODETYPE $WLCG_SERVICE);

my %MAP_SERVICETYPE_PORT = (
    BDII => 'BDII_PORT',
    ResourceBroker => 'RB_PORT',
    gsiftp => 'GRIDFTP_PORT',
    'MDS-GIIS' => 'MDS_PORT',
    'globus-GRAM' => 'GRAM_PORT',
    SRM => 'SRM1_PORT',
    DPNS => 'DPNS_PORT',
    'it.infn.GridICE' => 'GRIDICE_PORT',
    MyProxy => 'MYPROXY_PORT',
    'glite-LFC' => 'LFC_PORT',
    'org.glite.wms.NetworkServer' => 'WMS_PORT'
);

my $NPM_ROOT_URL = "https://ccenoc.in2p3.fr/DownCollector/";

# options
my ($OUTPUTFILE,$DEBUG,$VONAME,$NAGIOS_PLUGINS,$GLOBAL_SITES,$TEMPLS,$GLOBAL_SITE,$BDII,$MYPROXY,$MYPROXY_NAME,$PROBES,$PHASE);
my ($verbose, $WLCG_PLUGINS,$NO_BDII,$ADMINEMAIL,$SAM_INCLUDE_ALL);

my $COMMAND_LINE = "";

# sites:
#   sitename
#      -> hosts
#		-> hostname = nodetype
#		-> hostname = nodetype
#      -> contact 

###############################################################
#####
#####	Helper functions
#####
###############################################################

sub verbose {
    my $msg = shift;
    print $msg if $verbose;
}

sub exitError {
    my $msg = shift;
    print $msg;
    exit 1;
}

sub checkReadable {
    my $file = shift;
    ( -f $file) or exitError "Cannot find file $file\n";
    ( -r $file) or exitError "Cannot read $file\n";
}

sub checkWritable {
    my $file = shift;
    return if ! -f $file;
    (-f $file && -w $file) or exitError "Cannot write $file\n";
}

sub checkExecutable {
    my $file = shift;
    ( -f $file) or exitError "Cannot find file $file\n";
    ( -x $file) or exitError "Cannot execute $file\n";
}

sub checkArgs {
    my $site;
    my $help;

    # Store entire ARG string for later saving to generated config file
    foreach my $arg (@ARGV) {
	$COMMAND_LINE .= $arg." ";
    }

    if (!GetOptions (
        'output=s' => \$OUTPUTFILE, 
        'debug' => \$DEBUG, 
        'site=s' => \$GLOBAL_SITE, 
        'vo=s' => \$VONAME,
        'templates=s' => \$TEMPLS, 
        'h' => \$help, 'help' => \$help, 
        'v' => \$verbose, 'verbose' => \$verbose,
        'nagios=s' => \$NAGIOS_PLUGINS, 
        'bdii=s' => \$BDII, 
        'no-bdii' => \$NO_BDII,
        'myproxy=s' => \$MYPROXY,
        'myproxyname=s' => \$MYPROXY_NAME, 
        'probes=s' => \$PROBES,
        'phase=s' => \$PHASE,
        'admin=s' => \$ADMINEMAIL,
        'sam-all' => \$SAM_INCLUDE_ALL)) {
        exit 1;
    }

    if ($help) {
        print "        
Nagios configuration generator (WLCG probe based)

    Options:
        --output STRING
            Path to generated configuration file
            (default: /etc/nagios/wlcg.cfg)
        --site STRING
            Sitename.
            This option is mandatory.
        --vo STRING
            VO name 
            (default: dteam)

        --myproxy STRING
            Address of MyProxy server.
            This option is mandatory unless 'probes' is
            set to remote.
        --myproxyname STRING
            Name of credential stored in MyProxy server 
            (default: NagiosRefresh)

        --probes STRING
            Which probes to include in configuration.
            Possible values:
                local  - only locally executed probes are included. 
                         Nagios won't pull results from external
                         monitoring systems (SAM).
                remote - only remotely executed probes are included. 
                         Nagios won't run any active probes. MyProxy
                         settings are not required in this case.
                both   - both probe types are included.
            (default: both)
        --no-bdii
            Don't add new hosts found in BDII.
        --sam-all
            Include all SAM tests. Nagios will fetch critical only by default.

        --admin STRING
            List of email addresses set for Nagios internal checks:
            NPM-Gather, SAM-Gather, GridProxy-Get, GridProxy-Valid,
            MyProxy-ProxyLifetime, CADist-Version (on Nagios server).
            (default: root\@localhost)
            
        --templates STRING
            Path to templates directory 
            (default: /opt/lcg/share/grid-monitoring/config-gen/nagios)
        --nagios STRING
            Path to Nagios plugins directory
            (default: /usr/lib/nagios/plugins)
        --bdii STRING
            Address of bdii server. This option is necessary 
            only if BDII is not present in SAM db or 
            if the site is not in SAM at all.
                    
        -h, --help
            Basic program description

        -v, --verbose
            Print verbose logging information

    Examples:
        Generate full configuration for site egee.srce.hr:
            ./ncg.pl --site egee.srce.hr --myproxy myproxy.srce.hr --output /etc/nagios/wlcg.cfg

        Generate remote-only configuration for site egee.srce.hr:
            ./ncg.pl --site egee.srce.hr --probes remote

";
        exit 0;
    }

    $OUTPUTFILE or $OUTPUTFILE = '/etc/nagios/wlcg.cfg';
    $NAGIOS_PLUGINS or $NAGIOS_PLUGINS = '/usr/lib/nagios/plugins';
    $WLCG_PLUGINS or $WLCG_PLUGINS = '/opt/lcg/share/grid-monitoring/plugins';
    $GLOBAL_SITE or $GLOBAL_SITE = $SITE_NAME;
    $VONAME or $VONAME = 'dteam';
    $TEMPLS or $TEMPLS = '/opt/lcg/share/grid-monitoring/config-gen/nagios';
    $MYPROXY or $MYPROXY = $MYPROXY_SERVER;
    $MYPROXY_NAME or $MYPROXY_NAME = 'NagiosRefresh';
    $ADMINEMAIL or $ADMINEMAIL = 'root@localhost';

    # Check templates
    ( -d "$TEMPLS" ) or exitError "Cannot find directory with configuration templates. Set parameter templates.\n";
    checkReadable "$TEMPLS/contact.template";
    checkReadable "$TEMPLS/hostgroup.template";
    checkReadable "$TEMPLS/host.template";
    checkReadable "$TEMPLS/host-aggregate.template";
    checkReadable "$TEMPLS/servicegroup.template";
    checkReadable "$TEMPLS/servicedependency.template";
    checkReadable "$TEMPLS/master.template";
    checkReadable "$TEMPLS/myproxy.template";
    checkReadable "$TEMPLS/sam.template";
    checkReadable "$TEMPLS/gridproxy.template";
    checkReadable "$TEMPLS/cadist.template";
    checkReadable "$TEMPLS/npm.template";
    checkReadable "$TEMPLS/service/wlcg.template";
    checkReadable "$TEMPLS/service/native.template";
    checkReadable "$TEMPLS/service/remote.template";
    checkReadable "$TEMPLS/service/npm.template";

    # Check output file
    checkWritable $OUTPUTFILE;

    if ($PROBES) {
        if ($PROBES ne 'both' && $PROBES ne 'remote' && $PROBES ne 'local') {
            print "Incorrect probes parameter. Probes locality must be both, remote or local.\n";
            exit 1;
        }
    } else {
        $PROBES = 'both';
    }    

    exitError ("MyProxy server must be defined! Set parameter myproxy.\n") if ($PROBES ne 'remote' && !$MYPROXY);
    exitError ("Site must be defined! Set parameter site, or add variable 'SITE_NAME' to /etc/sysconfig/lcg.\n")
	if (!$GLOBAL_SITE);
    
    ( -d $NAGIOS_PLUGINS ) or exitError "Cannot find Nagios plugins directory. Set parameter nagios.\n";
    checkExecutable "$NAGIOS_PLUGINS/check_ping";
    if ($PROBES ne 'remote') {
        checkExecutable "$NAGIOS_PLUGINS/check_ftp";
        checkExecutable "$NAGIOS_PLUGINS/check_tcp";
    }
    verbose("Using:\n");
    verbose("  Output File         : $OUTPUTFILE\n");
    verbose("  Site                : $GLOBAL_SITE\n");
    verbose("  VO                  : $VONAME\n");
    verbose("  Probes              : $PROBES\n");
    verbose("  Nagios Plugins Dir  : $NAGIOS_PLUGINS\n");
    verbose("  MyProxy Server      : $MYPROXY\n") if $PROBES ne 'remote';

}

sub checkPing ($) {
    my $host = shift;
    my $retVal;

    `$NAGIOS_PLUGINS/check_ping -H $host -w 10000.000ms,100% -c 10000.000ms,100%`; 

    unless ($?) {
        $retVal = 1;
    } else {
        $retVal = 0;
    }

    $retVal;
}

sub checkCEGridFTP {
    my $host = shift;
    my $retVal;

    `$NAGIOS_PLUGINS/check_ftp -H $host -p 2811 -t 60`;

    unless ($?) {
        $retVal = 1;
    } else {
        $retVal = 0;
    }

    $retVal;
}

sub checkTcp {
    my $host = shift;
    my $port = shift;
    my $retVal;

    `$NAGIOS_PLUGINS/check_tcp -H $host -p $port -t 60`;

    unless ($?) {
        $retVal = 1;
    } else {
        $retVal = 0;
    }

    $retVal;
}

sub addHostAttributes {
    my $sitename = shift;
    my $hostname = shift;
    my $address = shift;

    if (!exists $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}) {
        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{sitename} = $sitename;
        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{hostname} = $hostname;
        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{notification} = 0;
        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{address} = $address;
        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{contactlist} = "$sitename-admins";
    }
}

sub addHost {
    my $sitename = shift;
    my $hostnamePublished = shift;
    my $noAdd = shift;
    
    my $address;
    my ($hostname, $hostname1, $aliases, $addrtype, $length, @addrs, @addrs1); 
    
    # get detailed information about this host & 
    # check if host exists at all
    unless ( ($hostname, $aliases, $addrtype, $length, @addrs) = gethostbyname($hostnamePublished)) {
        return undef;
    }

    # id noAdd is set, don't add the host, just return its real name
    # if the host exists
    if ($noAdd) {
        if (exists $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}) {
            return $hostname;
        } else {
            return undef;
        }
    }

    # if we're dealing with remote-only setup don't generate
    # config for real hostnames & load-balanced nodes        
    if ($PROBES eq 'remote') {
        $hostname = $hostnamePublished;

        # use hostname instead of IP address for address field if
        # host is alias or has load-balanced nodes
        if ($#addrs || $aliases) {
            $address = $hostname;
        } else {
            $address = join ('.', unpack('C4', $addrs[0]));
        }
        
        addHostAttributes($sitename, $hostname, $address);
    } else {
        # here we use hostname for address field only if host has 
        # load-balanced nodes
        if ($#addrs) {
            $address = $hostname;
        } else {
            $address = join ('.', unpack('C4', $addrs[0]));
        }

        addHostAttributes($sitename, $hostname, $address);
        
        # if hostname is alias add it to list of aliases
        # use alias for address field
        if ($aliases) {
            $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{aliases}->{$hostnamePublished}->{address} = $hostnamePublished;
            $GLOBAL_SITES->{$sitename}->{aliases}->{$hostnamePublished} = $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{aliases}->{$hostnamePublished};
        } 
        
        # if there are load-balanced nodes add them to list
        # of aliases, use their IP addresses for address field
        if ($#addrs) {
            foreach my $addr (@addrs) {
                if (($hostname1, $aliases, $addrtype, $length, @addrs1) = gethostbyaddr($addr, AF_INET)) {
                    $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{lbnodes}->{$hostname1}->{address} = join ('.', unpack('C4', $addr));    
                }
            }
        }
    }

    $hostname;
}


###############################################################
#####
#####   Hosts analyzers
#####
###############################################################


###############################################################
#####
#####   SAM information gatherer
#####
###############################################################
sub transformSAMName ($) {
    my $samname = shift;
    if ($samname eq 'RGMA') {
        $samname = 'MON';
    } elsif ($samname eq 'MyProxy') {
        $samname = 'PROX';
    } elsif ($samname eq 'sBDII') {
        $samname = 'Site-BDII';
    } elsif ($samname eq 'gCE') {
        $samname = 'gLite-CE';
    } elsif ($samname eq 'gRB') {
        $samname = 'WMS';
    }

    $samname;
}

sub inverseSAMName ($) {
    my $samname = shift;
    if ($samname eq 'MON') {
        $samname = 'RGMA';
    } elsif ($samname eq 'PROX') {
        $samname = 'MyProxy';
    } elsif ($samname eq 'Site-BDII') {
        $samname = 'sBDII';
    } elsif ($samname eq 'gLite-CE') {
        $samname = 'gCE';
    } elsif ($samname eq 'WMS') {
        $samname = 'gRB';
    }
    $samname;
}

my $XML_SITE_NAME = "";

sub xmlGetInfoSAMStart
{
    my $p = shift;
    my $data = shift;
    my $name;
    my $value;
    my $hostname;
    my $nodetype;

    if ($p->eq_name($data, "Service") )
    {
        while ($name = shift) {
            $value = shift;
            if ($name eq 'endpoint') {
                $hostname = $value;
            } elsif ($name eq 'type') {
                $nodetype = transformSAMName($value);
                $nodetype = undef if (!exists $WLCG_NODETYPE->{$nodetype});
            }
        }

        if ($hostname && $nodetype) {
            $hostname = addHost($XML_SITE_NAME, $hostname);
            return if (!$hostname);
            $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$hostname}->{nodetypeArr}->{$nodetype} = {} unless (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$hostname}->{nodetypeArr}->{$nodetype});
            $GLOBAL_SITES->{$XML_SITE_NAME}->{bdii} = $hostname if ($nodetype eq 'Site-BDII');
        }
    }
}

sub getInfoSAM {
    my $sitename    = shift;

    my $ua = LWP::UserAgent->new;
    $ua->agent("Nagios config generator/0.2 ");

    # Create a request
    verbose("Getting info from SAM for site $sitename:\n");
    my $req = HTTP::Request->new(POST => 'http://lcg-sam.cern.ch:8080/same-pi/service_endpoint_status.jsp?VO_name=OPS&Site_name='.$sitename.'&show_details');
    # Pass request to the user agent and get a response back
    my $res = $ua->request($req);

    # Check the outcome of the response
    if (!$res->is_success) {
        print $res->status_line, "\n";
        exitError("Could not get results from SAM : ".$res->status_line);
    }
    
    $GLOBAL_SITES->{$sitename}->{sitename} = $sitename;
    $XML_SITE_NAME = $sitename;
    
    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Start => \&xmlGetInfoSAMStart);
    $parser->parse($res->content);

    $XML_SITE_NAME = "";

    verbose("  Have ". keys (%{$GLOBAL_SITES->{$sitename}->{hosts}})." hosts.\n");
	verbose("    ".join("\n    ",sort keys(%{$GLOBAL_SITES->{$sitename}->{hosts}}))."\n");
    
    verbose("Done!\n");
    1;
}

###############################################################
#####
#####   BDII information gatherer
#####
###############################################################

sub getInfoBDII {
    my $sitename = shift;
    my $bdii = shift;
    my $vo = shift || "dteam";
    my $basedn = "Mds-Vo-Name=$sitename,O=Grid";
    my $mesg;
    my $hostname;
    my $port;
    my $entry;
    my $sitehosts = $GLOBAL_SITES->{$sitename}->{hosts};

    verbose("Getting info from BDII $bdii for site $sitename:\n");
    my $ldap = Net::LDAP->new( $bdii, port => 2170, timeout => 60, async => 1 );
    # bind to a directory with dn and password
    $mesg = $ldap->bind( anonymous => 1 );
    $mesg->code && return;

    # Get site contact address
    unless (defined $GLOBAL_SITES->{$sitename}->{contact}) {
        $mesg = $ldap->search ( base => "Mds-Vo-Name=$sitename,O=Grid", filter => "(objectClass=GlueSite)", attrs=> ['GlueSiteSysAdminContact'], timelimit => 60 );
        $GLOBAL_SITES->{$sitename}->{contact} = $1 if (!$mesg->code && $mesg->count != 0 && $mesg->entry(0)->get_value('GlueSiteSysAdminContact') =~ /^mailto:\s*(\S+)/);
    }

    if ($PROBES eq 'remote') {
        verbose("  Have contact address: ". $GLOBAL_SITES->{$sitename}->{contact} .".\n");
        verbose("Done!\n");
        return;
    }

    # Get SE information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(objectClass=GlueSE)",
                            attrs=> ['GlueSEUniqueID'],
                            timelimit => 60 );
    unless ($mesg->code){
        foreach $entry ($mesg->entries()) {
            my $hostnamePublished = $entry->get_value('GlueSEUniqueID');
            $hostname = addHost($sitename, $hostnamePublished, $NO_BDII);

            # skip further analysis if
            #  - hostaddress was incorrect or
            #  - nobdii is switched on & host is not already in the list
            next if (!$hostname);

            # Get remote path that selected VO can use
            my $mesg1 = $ldap->search ( base => $basedn,
                                        filter => "(&(&(objectClass=GlueSA)(GlueSAAccessControlBaseRule=$vo))(GlueChunkKey=GlueSEUniqueID=$hostnamePublished))",
                                        attrs => ['GlueSAPath'],
                                        timelimit => 60 );
            next if ($mesg1->code);
            if ($mesg1->count != 0) {
                $sitehosts->{$hostname}->{SE_PATH} = $mesg1->entry(0)->get_value('GlueSAPath');
            }
            $sitehosts->{$hostname}->{nodetypeArr}->{SE} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{SE});
            if ($sitehosts->{$hostname}->{SE_PATH} && $sitehosts->{$hostname}->{SE_PATH} =~ /^\/dpm\//) {
                $sitehosts->{$hostname}->{nodetypeArr}->{DPM} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{DPM});
            }

            # GridFTP specific parameter
            $mesg1 = $ldap->search ( base => $basedn,
                                        filter => "(&(&(objectClass=GlueSEAccessProtocol)(GlueChunkKey=GlueSEUniqueID=$hostnamePublished))(GlueSEAccessProtocolType=gsiftp))",
                                        attrs => ['GlueSEAccessProtocolPort'],
                                        timelimit => 60 );
            next if ($mesg1->code);
            if ($mesg1->count != 0) {
                $sitehosts->{$hostname}->{GRIDFTP_PORT} = $mesg1->entry(0)->get_value('GlueSEAccessProtocolPort');
                $sitehosts->{$hostname}->{services}->{gsiftp}->{serviceuri} = $hostname . ":" . $sitehosts->{$hostname}->{GRIDFTP_PORT};

            } else {
                $sitehosts->{$hostname}->{missingservices}->{gsiftp} = {};
            }
        }
    }

    # Get CE information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(objectClass=GlueCE)(GlueCEAccessControlBaseRule=VO:$vo))",
                            attrs=> ['GlueCEUniqueID'],
                            timelimit => 60 );
    unless ($mesg->code){
        foreach $entry ($mesg->entries()) {
            #lxb2006.cern.ch:2119/jobmanager-lcgpbs-dteam
            #lxb2039.cern.ch:2119/blah-pbs-dteam
            if ($entry->get_value('GlueCEUniqueID') =~ /([-_.A-Za-z0-9]+):(\d+)\/(jobmanager|blahp)-([-_.A-Za-z0-9]+)-([-_.A-Za-z0-9]+)/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);

                # skip further analysis if
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list
                next if (!$hostname);

                $sitehosts->{$hostname}->{nodetypeArr}->{CE} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{CE});
                $sitehosts->{$hostname}->{GRAM_PORT} = $port;
                $sitehosts->{$hostname}->{services}->{'globus-GRAM'}->{serviceuri} = $hostname . ":" . $port;
            }
        }
    }

    # Get SRM information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(&(objectClass=GlueService)(GlueServiceType=srm_v1))(GlueServiceAccessControlRule=$vo))",
                            attrs=> ['GlueServiceURI'],
                            timelimit => 60 );
    unless ($mesg->code) { 
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceURI') =~ /httpg:\/\/([-_.A-Za-z0-9]+):(\d+)\/srm\/managerv1/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);

                # skip further analysis if 
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                
                next if (!$hostname);

                $sitehosts->{$hostname}->{nodetypeArr}->{SRM} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{SRM});
                $sitehosts->{$hostname}->{SRM1_PORT} = $port;
                $sitehosts->{$hostname}->{services}->{SRM}->{serviceuri} = "httpg://$hostname:$port/srm/managerv1";
            }
        }
    }

    # Get SRM2 information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(&(objectClass=GlueService)(GlueServiceType=srm))(GlueServiceAccessControlRule=$vo))",
                            attrs=> ['GlueServiceURI'],
                            timelimit => 60 );
    unless ($mesg->code) {
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceURI') =~ /httpg:\/\/([-_.A-Za-z0-9]+):(\d+)\/srm\/managerv2/) {
                $hostname = $1;
                $port = $2;
                
                # 1 is set here because we don't want to add pure SRM2 nodes
                $hostname = addHost($sitename, $hostname, 1);

                # skip further analysis if
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                
                next if (!$hostname);
                
                # don't check GridFTP on SRM2 nodes
                unless (checkCEGridFTP($hostname)) {
                    $sitehosts->{$hostname}->{missingservices}->{gsiftp} = {};
                }
            }
        }
    }


    # Get Resource Broker information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(&(objectClass=GlueService)(GlueServiceType=ResourceBroker))(GlueServiceAccessControlRule=$vo))",
                            attrs=> ['GlueServiceEndpoint'],
                            timelimit => 60 );
    unless ($mesg->code) {
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceEndpoint') =~ /([-_.A-Za-z0-9]+):(\d+)/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);
                
                # skip further analysis if 
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                      
                next if (!$hostname);

                $sitehosts->{$hostname}->{nodetypeArr}->{RB} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{RB});
                $sitehosts->{$hostname}->{RB_PORT} = $port;
                #$sitehosts->{$hostname}->{services}->{ResourceBroker}->{serviceuri} = $entry->get_value('GlueServiceEndpoint');
                $sitehosts->{$hostname}->{services}->{ResourceBroker}->{serviceuri} = "$hostname:$port";
            }
        }
    }

    # Get Resource Broker information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(&(objectClass=GlueService)(GlueServiceType=org.glite.wms))(GlueServiceAccessControlRule=$vo))",
                            attrs=> ['GlueServiceUniqueID'],
                            timelimit => 60 );
    unless ($mesg->code) {
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceUniqueID') =~ /([-_.A-Za-z0-9]+):(\d+)/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);
                
                # skip further analysis if
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                  
                next if (!$hostname);
                
                $sitehosts->{$hostname}->{nodetypeArr}->{WMS} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{WMS});
                $sitehosts->{$hostname}->{WMS_PORT} = $port;
                #$sitehosts->{$hostname}->{services}->{'org.glite.wms.NetworkServer'}->{serviceuri} = $entry->get_value('GlueServiceUniqueID');
                $sitehosts->{$hostname}->{services}->{'org.glite.wms.NetworkServer'}->{serviceuri} = "$hostname:$port";
            }
        }
    }
    

    # Get MyProxy information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(objectClass=GlueService)(GlueServiceType=MyProxy))",
                            attrs=> ['GlueServiceEndpoint'],
                            timelimit => 60 );
    unless ($mesg->code) {
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceEndpoint') =~ /([-_.A-Za-z0-9]+):(\d+)/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);
                
                # skip further analysis if
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                  
                next if (!$hostname);
                
                $sitehosts->{$hostname}->{nodetypeArr}->{PROX} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{PROX});
                $sitehosts->{$hostname}->{MYPROXY_PORT} = $port;
                #$sitehosts->{$hostname}->{services}->{MyProxy}->{serviceuri} = $entry->get_value('GlueServiceEndpoint');
                $sitehosts->{$hostname}->{services}->{MyProxy}->{serviceuri} = "$hostname:$port";
            }
        }
    }

    # Get MON boxes information
    $mesg = $ldap->search ( base => $basedn,
                            filter => "(&(objectClass=GlueService)(GlueServiceType=gridice))",
                            attrs=> ['GlueServiceUniqueID'],
                            timelimit => 60 );
    unless ($mesg->code) {
        foreach $entry ($mesg->entries()) {
            if ($entry->get_value('GlueServiceUniqueID') =~ /([-_.A-Za-z0-9]+):(\d+)/) {
                $hostname = $1;
                $port = $2;
                $hostname = addHost($sitename, $hostname, $NO_BDII);

                # skip further analysis if 
                #  - hostaddress was incorrect or
                #  - nobdii is switched on & host is not already in the list                  
                next if (!$hostname);
                
                $sitehosts->{$hostname}->{nodetypeArr}->{MON} = {} unless (exists $sitehosts->{$hostname}->{nodetypeArr}->{MON});
                $sitehosts->{$hostname}->{GRIDICE_PORT} = $port;
            }
        }
    }
    verbose("  Have ". keys (%{$GLOBAL_SITES->{$sitename}->{hosts}})." hosts.\n");
	verbose("    ".join("\n    ",sort keys(%{$GLOBAL_SITES->{$sitename}->{hosts}}))."\n");
    verbose("Done!\n");
    1;
}

###############################################################
#####
#####   Service analyzers
#####
###############################################################



###############################################################
#####
#####   SAM remote services gatherer
#####
###############################################################


#####################################################################
###
###  Get list of supported VOs per site
###
#####################################################################

my $XML_LIST_OF_SUPPORTED_VOS;
my $XML_VO_NAME;
my $XML_REMOTE_HOST = "";
my $XML_REMOTE_SERVICE = "";
my $XML_REMOTE_SERVICES = {};

sub xmlGetSupportedVOsChar
{
    my $p = shift;
    my $data = shift;

    if ($p->in_element("supportsVO")) {
        $XML_LIST_OF_SUPPORTED_VOS->{$data} = 1;
    }
}

sub getSupportedVOs {
    my $sitename = shift;
    my $ua = LWP::UserAgent->new;
    $ua->agent("Nagios config generator/0.2 ");

    my $req = HTTP::Request->new(POST => 'http://lcg-sam.cern.ch:8080/same-pi/services_per_vo.jsp?Site_name='.$sitename);
    my $res = $ua->request($req);

    if (!$res->is_success) {
        return (0, "SAM gathering failed with HTTP error: " . $res->status_line . "\n");
    }

    $XML_LIST_OF_SUPPORTED_VOS={};

    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Char => \&xmlGetSupportedVOsChar);
    $parser->parse($res->content);

    return (1,keys %{$XML_LIST_OF_SUPPORTED_VOS});
}
#####################################################################


#####################################################################
###
###  Get critical checks for defined VO
###
#####################################################################


sub xmlGetRemoteCriticalServiceStart
{
    my $p = shift;
    my $data = shift;
    my $name;
    my $value;

    if ($p->eq_name($data, 'ServiceMetric') )
    {
        $name = shift;
        $value = shift;
        if ($name eq 'name') {
            $XML_REMOTE_SERVICES->{$XML_VO_NAME}->{$value . '-' . $XML_VO_NAME} = 1;
        }
    }
}

sub getCriticalsVO ($) {
    my $vo = shift;

    return  (1, $XML_REMOTE_SERVICES) if (exists $XML_REMOTE_SERVICES->{$vo});

    my $ua = LWP::UserAgent->new;
    my $status;
    my $result;

    $ua->agent("Nagios config generator/0.2 ");

    my $serviceTypesStr = '';

    foreach my $serviceType (keys %$WLCG_NODETYPE) {
        $serviceTypesStr .= '&Service_type=' . inverseSAMName($serviceType);
    }

    my $req = HTTP::Request->new(POST => 'http://lcg-sam.cern.ch:8080/same-pi/servicemetrics_per_service_type_critical_for_vo.jsp?VO_name='.$vo.$serviceTypesStr);
    my $res = $ua->request($req);
    if (!$res->is_success) {
        return (0, "Gathering SAM critical checks failed with HTTP error: " . $res->status_line . "\n");
    }

    $XML_VO_NAME = $vo;

    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Start => \&xmlGetRemoteCriticalServiceStart);
    $parser->parse($res->content);

    return (1, $XML_REMOTE_SERVICES);
}

#####################################################################

#####################################################################
###
###  Get all services on a site
###
#####################################################################


sub xmlGetSAMRemoteServiceStart
{
    my $p = shift;
    my $data = shift;
    my $name;
    my $value;

    if ($p->eq_name($data, 'Service') )
    {
        $name = shift;
        $value = shift;
        if ($name eq 'endpoint') {
            $XML_REMOTE_HOST = $value;
        }
    }
    if ($p->eq_name($data, 'ServiceMetric') )
    {
        $name = shift;
        $value = shift;
        if ($name eq 'name') {
            $XML_REMOTE_SERVICE = $value . '-' . $XML_VO_NAME;
        }
        if (!$SAM_INCLUDE_ALL) {
            $XML_REMOTE_SERVICE = undef if (! exists $XML_REMOTE_SERVICES->{$XML_VO_NAME}->{$XML_REMOTE_SERVICE});
        }
    }
}

sub xmlGetSAMRemoteServiceChar
{
    my $p = shift;
    my $data = shift;

    if ($XML_REMOTE_HOST && $XML_REMOTE_SERVICE) {
        if ($p->in_element("DETAILED_URL")) {
            if (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_REMOTE_HOST}) {
                if (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}) {
                    $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}->{url} .= $data;
                } else {
                    $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}->{url} = $data;
                } 
            } elsif (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_REMOTE_HOST}) {
                if (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}) {
                    $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}->{url} .= $data;
                } else {
                    $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_REMOTE_HOST}->{remoteservices}->{$XML_REMOTE_SERVICE}->{url} .= $data; 
                }
            }

            #$XML_REMOTE_SERVICES->{$XML_REMOTE_HOST}->{$XML_REMOTE_SERVICE}->{url} = $data;
        }
    }
}

sub xmlGetSAMRemoteServiceEnd
{
    my $p = shift;
    my $data = shift;

    if ($p->eq_name($data, "Service") )
    {
        $XML_REMOTE_HOST = "";
    }
    if ($p->eq_name($data, "ServiceMetric") )
    {
        $XML_REMOTE_SERVICE = "";
    }
}

sub getRemoteServicesSiteVO ($$) {
    my $sitename = shift;
    my $vo = shift;
    my $ua = LWP::UserAgent->new;
    $ua->agent("Nagios config generator/0.2 ");

    my $req = HTTP::Request->new(POST => 'http://lcg-sam.cern.ch:8080/same-pi/service_endpoint_status.jsp?VO_name='.$vo.'&Site_name='.$sitename.'&show_details');
    my $res = $ua->request($req);

    if (!$res->is_success) {
        return (0, "SAM gathering failed with HTTP error: " . $res->status_line . "\n");
    }

    if (!$SAM_INCLUDE_ALL) {
        my $output;
        my $status;
        ( $status, $output ) = getCriticalsVO($vo);
        return ($status, $output) if  ( !$status );
    }

    $XML_VO_NAME = $vo;


    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Start => \&xmlGetSAMRemoteServiceStart, End => \&xmlGetSAMRemoteServiceEnd, Char => \&xmlGetSAMRemoteServiceChar);
    $parser->parse($res->content);

    # fill remote services
    #foreach my $hostname (keys %{$XML_REMOTE_SERVICES})
    #{
    #    if (exists $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}) {
    #        $GLOBAL_SITES->{$sitename}->{hosts}->{$hostname}->{remoteservices} = {%{$XML_REMOTE_SERVICES->{$hostname}}};
    #    } elsif (exists $GLOBAL_SITES->{$sitename}->{aliases}->{$hostname}) {
    #        $GLOBAL_SITES->{$sitename}->{aliases}->{$hostname}->{remoteservices} = {%{$XML_REMOTE_SERVICES->{$hostname}}};
    #    }
    #}

    1;
}

sub getSAMRemoteServices {
    my $sitename = shift;
    my $status;
    my @VOs;

    $XML_SITE_NAME = $sitename;

    ($status, @VOs) = getSupportedVOs($sitename);
    if (!$status) {
        return;
    }

    foreach my $vo (@VOs) {
        return 0 unless (getRemoteServicesSiteVO ($sitename, $vo));
    }

    1;
}

###############################################################
#####
#####   NPM remote services gatherer
#####
###############################################################

my $XML_NPM_CURRENT_HOST = "";
my $XML_NPM_CURRENT_TYPE = "";

sub xmlGetNPMRemoteServiceStart
{
    my $p = shift;
    my $data = shift;
    my $name;
    my $value;

    if ($p->eq_name($data, 'Service') )
    {
        while ($name = shift) {
            $value = shift;
            if ($name eq 'endpoint') {
                if ($value =~ /([-_.A-Za-z0-9]+):(\d+)/) {
                   $XML_NPM_CURRENT_HOST = $1;
                }
            } elsif ($name eq 'type') {
                $XML_NPM_CURRENT_TYPE = $value;
            }
        }
    } elsif ($p->eq_name($data, 'ServiceMetric')) {
        $name = shift;
        $value = shift;
        if ($name eq 'name' && $XML_NPM_CURRENT_HOST && $XML_NPM_CURRENT_TYPE) {
            if (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_NPM_CURRENT_HOST}) {
                $GLOBAL_SITES->{$XML_SITE_NAME}->{hosts}->{$XML_NPM_CURRENT_HOST}->{npmservices}->{"org.egee.npm.".$XML_NPM_CURRENT_TYPE}->{url} = $NPM_ROOT_URL."?nd=$XML_NPM_CURRENT_TYPE,$XML_NPM_CURRENT_HOST";
            } elsif (exists $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_NPM_CURRENT_HOST}) {
                $GLOBAL_SITES->{$XML_SITE_NAME}->{aliases}->{$XML_NPM_CURRENT_HOST}->{npmservices}->{"org.egee.npm.".$XML_NPM_CURRENT_TYPE}->{url} = $NPM_ROOT_URL."?nd=$XML_NPM_CURRENT_TYPE,$XML_NPM_CURRENT_HOST";
            }
        }
    }
}

sub xmlGetNPMRemoteServiceEnd
{
    my $p = shift;
    my $data = shift;

    if ($p->eq_name($data, 'Service') )
    {
        $XML_NPM_CURRENT_HOST = "";
        $XML_NPM_CURRENT_TYPE = "";
    }
}

sub getNPMRemoteServices {
    my $sitename = shift;
    my $status;

    $XML_SITE_NAME = $sitename;

    my $ua = LWP::UserAgent->new;
    $ua->agent("Nagios config generator/0.2 ");

    my $req = HTTP::Request->new(POST => $NPM_ROOT_URL.'EMS/EMSMonitoringService.php?proc=getAllMetrics&sitename='.$XML_SITE_NAME);
    my $res = $ua->request($req);

    if (!$res->is_success) {
        return (0, "NPM gathering failed with HTTP error: " . $res->status_line . "\n");
    }

    my $parser = new XML::Parser(ErrorContext => 2);
    $parser->setHandlers(Start => \&xmlGetNPMRemoteServiceStart, End => \&xmlGetNPMRemoteServiceEnd );
    $parser->parse($res->content);

    1;
}

###############################################################


sub getRemoteServices {
    my $sitename = shift;
    
    getSAMRemoteServices $sitename;
    
    getNPMRemoteServices $sitename;
}


###############################################################
#####
#####   Probe description DB services
#####
###############################################################

sub setHostCheckType {
    my $node = shift;
    if ( checkPing($node->{address}) ) {
        $node->{checktype} = "check-host-alive";
    } else {
        if (exists $node->{nodetypeArr}->{CE} || exists $node->{nodetypeArr}->{'gLite-CE'}) {
            $node->{checktype} = "check-tcp!2119";
        } elsif (exists $node->{nodetypeArr}->{SE}) {
            $node->{checktype} = "check-tcp!2811";
        } elsif (exists $node->{nodetypeArr}->{'Site-BDII'} || exists $node->{nodetypeArr}->{BDII}) {
            $node->{checktype} = "check-tcp!2170";
        } elsif (exists $node->{nodetypeArr}->{WMS} || exists $node->{nodetypeArr}->{RB}) {
            $node->{checktype} = "check-tcp!7772";
        } elsif (exists $node->{nodetypeArr}->{LFC}) {
            $node->{checktype} = "check-tcp!5010";
        } elsif (exists $node->{nodetypeArr}->{PROX}) {
            $node->{checktype} = "check-tcp!7512";
        } elsif (exists $node->{nodetypeArr}->{MON} || exists $node->{nodetypeArr}->{VOMS}) {
            $node->{checktype} = "check-tcp!8443";
        }
    }
}

sub getServices {
    verbose("Getting service info with live checks:\n");

    foreach my $site (keys %$GLOBAL_SITES){
        foreach my $node (values %{$GLOBAL_SITES->{$site}->{hosts}}) {
            my $sitename = $node->{sitename};
            my $hostname = $node->{hostname};
            verbose("  Analyzing host: ".$hostname."\n");

            foreach my $nodeT (keys %{$node->{nodetypeArr}}) {
                foreach my $service (@{$WLCG_NODETYPE->{$nodeT}->{services}}) {
                    $node->{services}->{$service} = {} unless (exists $node->{services}->{$service});
                }
            }

            $node->{nodetype} = join (",", keys %{$node->{nodetypeArr}});

            ### Dynamic information generation
            # check which method to use for checking hosts
            setHostCheckType($node);

            # GridFTP is closed on CE
            if (exists $node->{nodetypeArr}->{CE} || exists $node->{nodetypeArr}->{'gLite-CE'}) {
                unless (checkCEGridFTP($node->{address})) {
                    delete $node->{services}->{gsiftp};
                }
            }

            # Check if SE has GridFTP interface
            if (exists $node->{nodetypeArr}->{SE}) {
                if (exists $node->{missingservices}->{gsiftp}) {
                    delete $node->{services}->{gsiftp};
                }
            }

            # Analyze MON box
            if (exists $node->{nodetypeArr}->{MON}) {
                if (exists $node->{GRIDICE_PORT}) {
                    $node->{GRIDICE_DN} = "Mds-Vo-Name=local,O=Grid";
                } elsif (checkTcp($hostname,2136)) {
                    $node->{GRIDICE_PORT} = 2136;
                    $node->{GRIDICE_DN} = "Mds-Vo-Name=local,O=Grid";
                } else {
                    delete $node->{services}->{'it.infn.GridICE'};
                }
            }
            
            # Analyze SRM Box - GridFTP is not on this host
            if (exists $node->{nodetypeArr}->{SRM}) {
                unless (checkCEGridFTP($node->{address})) {
                    delete $node->{services}->{gsiftp};
                } 
            }

            # Set default config attributes (in case they're not extracted from BDII)

            # Set GridFTP port needed for native checks
            if (exists $node->{services}->{gsiftp}) {
                if (! exists $node->{GRIDFTP_PORT}) {
                    $node->{GRIDFTP_PORT} = 2811;
                }
                if (! exists $node->{SE_PATH}) {
                    $node->{SE_PATH} = "/tmp";
                }
            }

            # Set RB port if not extracted from BDII
            if (exists $node->{services}->{'globus-GRAM'} && ! exists $node->{GRAM_PORT}) {
                $node->{GRAM_PORT} = 2119;
            }

            # Set RB port if not extracted from BDII
            if (exists $node->{services}->{ResourceBroker} && ! exists $node->{RB_PORT}) {
                $node->{RB_PORT} = 7772;
            }

            # Set WMS port if not extracted from BDII
            if (exists $node->{services}->{'org.glite.wms.NetworkServer'} && ! exists $node->{WMS_PORT}) {
                $node->{WMS_PORT} = 7772;
            }

            # Set MyProxy port if not extracted from BDII
            if (exists $node->{services}->{MyProxy} && ! exists $node->{MYPROXY_PORT}) {
                $node->{MYPROXY_PORT} = 7512;
            }

            # Set SRM1 data if not extracted from BDII
            if (exists $node->{services}->{SRM}) {
                if (! exists $node->{SRM1_PORT}) {
                    $node->{SRM1_PORT} = 8443;
                }

                if (! exists $node->{SE_PATH}) {
                    $node->{SE_PATH} = "/tmp";
                }
            }

            # Set LDAP check parameter for BDII
            if (exists $node->{services}->{BDII}) {
                $node->{BDII_PORT} = 2170;
                if (exists $node->{nodetypeArr}->{BDII}) {
                    $node->{BDII_DN} = "Mds-Vo-Name=local,O=Grid";
                } elsif (exists $node->{nodetypeArr}->{'Site-BDII'}){
                    $node->{BDII_DN} = "GlueSiteUniqueID=$sitename,Mds-Vo-Name=$sitename,O=Grid";
                } elsif (exists $node->{nodetypeArr}->{DPM}){
                    $node->{BDII_DN} = "GlueSEUniqueID=$hostname,Mds-Vo-Name=resource,O=Grid";
                }
            }

            # Check if GRIS is opened & set LDAP check parameter for specific node type
            if (exists $node->{services}->{'MDS-GIIS'}) {
                $node->{MDS_PORT} = 2135;
                if (checkTcp($hostname,$node->{MDS_PORT})) {
                    if (exists $node->{nodetypeArr}->{CE} || exists $node->{nodetypeArr}->{'gLite-CE'}) {
                        $node->{MDS_DN} = "GlueClusterUniqueID=$hostname,Mds-Vo-Name=local,O=Grid";
                    } elsif (exists $node->{nodetypeArr}->{SE}) {
                        $node->{MDS_DN} = "GlueSEUniqueID=$hostname,Mds-Vo-Name=local,O=Grid";
                    } elsif (exists $node->{nodetypeArr}->{RB}) {
                        $node->{MDS_DN} = "GlueServiceUniqueID=$hostname:$node->{RB_PORT},Mds-Vo-Name=local,O=Grid";
                    } elsif (exists $node->{nodetypeArr}->{LFC}) {
                        $node->{MDS_DN} = "GlueServiceUniqueID=$hostname,Mds-Vo-Name=local,O=Grid";
                    } elsif (exists $node->{nodetypeArr}->{PROX}) {
                        $node->{MDS_DN} = "GlueServiceUniqueID=$hostname:$node->{MYPROXY_PORT},mds-vo-name=local,o=grid";
                    } elsif (exists $node->{services}->{'it.infn.GridICE'}) {
                        $node->{MDS_DN} = "GlueServiceUniqueID=$hostname:$node->{GRIDICE_PORT},mds-vo-name=local,o=grid";
                    } else {
                        $node->{MDS_DN} = "mds-vo-name=local,o=grid";
                    }
                } else {
                    delete $node->{services}->{'MDS-GIIS'};
                }
            }

            if (exists $node->{services}->{CAdistribution}) {
                if (exists $node->{services}->{gsiftp}) {
                    $node->{services}->{CAdistribution}->{serviceuri} = $hostname.":".$node->{GRIDFTP_PORT};
                } else {
                    delete $node->{services}->{CAdistribution};
                }
            }

            # Set serviceURI for all services
            foreach my $service (keys %{$node->{services}}) {
                if (! $node->{services}->{$service}->{serviceuri}) {
                    if (exists $MAP_SERVICETYPE_PORT{$service} && exists $node->{$MAP_SERVICETYPE_PORT{$service}}) {
                        $node->{services}->{$service}->{serviceuri} = $hostname . ':' . $node->{$MAP_SERVICETYPE_PORT{$service}};
                    } else {
                        $node->{services}->{$service}->{serviceuri} = $hostname;
                    }
                }
            }

        }
        getRemoteServices $site if ($PROBES ne 'local');
    }
    
    verbose("Done!\n");
}

sub getRemoteServicesOnly {
    verbose("Getting service info from SAM:\n");
    foreach my $site (keys %$GLOBAL_SITES){
        foreach my $node (values %{$GLOBAL_SITES->{$site}->{hosts}}) {
            my $sitename = $node->{sitename};
            my $hostname = $node->{hostname};

            $node->{nodetype} = join (",", keys %{$node->{nodetypeArr}});

            setHostCheckType($node);
        }

        getRemoteServices $site;
    }
    verbose("Done!\n");
}

###############################################################
#####
#####   Config generators
#####
###############################################################

sub genGridProxy {
    my $vo = shift;
    my $myproxy = shift;
    my $name = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    
    open ($TEMPL, "$TEMPLS/gridproxy.template");
    while ($line = <$TEMPL>){
        $line =~ s/<VONAME>/$vo/g;
        $line =~ s/<MYPROXY>/$myproxy/g;
        $line =~ s/<MYPROXY_NAME>/$name/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genMyProxy {
    my $name = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    
    open ($TEMPL, "$TEMPLS/myproxy.template");
    while ($line = <$TEMPL>){
        $line =~ s/<MYPROXY_NAME>/$name/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genRemote {
    my $sitename = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    my $samall = "";

    $samall = "--sam-all" if ($SAM_INCLUDE_ALL);

    open ($TEMPL, "$TEMPLS/sam.template");
    while ($line = <$TEMPL>){
        $line =~ s/<sitename>/$sitename/g;
        $line =~ s/<samall>/$samall/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
    
    open ($TEMPL, "$TEMPLS/npm.template");
    while ($line = <$TEMPL>){
        $line =~ s/<sitename>/$sitename/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genCADist {
    my $CONFIG = shift;
    my $TEMPL;
    my $line;

    open ($TEMPL, "$TEMPLS/cadist.template");
    while ($line = <$TEMPL>){
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genContact {
    my $site = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    my $contact = $site->{contact} || 'root@localhost';

    open ($TEMPL, "$TEMPLS/contact.template");
    while ($line = <$TEMPL>){
        $line =~ s/<sitename>/$site->{sitename}/g;
        $line =~ s/<siteemail>/$contact/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genHostGroup {
    my $name = shift;
    my $fullname = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    
    open ($TEMPL, "$TEMPLS/hostgroup.template");
    while ($line = <$TEMPL>){
        $line =~ s/<name>/$name/g;
        $line =~ s/<fullname>/$fullname/g;
        print $CONFIG $line;
    }
    close ($TEMPL);
}

sub genHostGroupSiteNode {
    my $sitename = shift;
    my $CONFIG = shift;

    genHostGroup ("site-$sitename", "Site $sitename", $CONFIG);
}

sub genHostGroupNodeType {
    my $nodename = shift;
    my $CONFIG = shift;

    genHostGroup ("node-$nodename", "EGEE $nodename", $CONFIG);
}

sub genHostGroupAliases {
    my $hostname = shift;
    my $CONFIG = shift;

    genHostGroup ("aliases-$hostname", "Aliases of host $hostname", $CONFIG);
    
    return "aliases-$hostname";
}

sub genHostGroupLBNodes {
    my $hostname = shift;
    my $CONFIG = shift;

    genHostGroup ("lbnodes-$hostname", "Load balanced nodes of host $hostname", $CONFIG);
    
    return "lbnodes-$hostname";
}

sub genHostAggregate {
    my $hostname = shift;
    my $contactlist = shift;
    my $servicegroup = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;

    open ($TEMPL, "$TEMPLS/host-aggregate.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<hostname>/$hostname/mg;
    $line =~ s/<contactlist>/$contactlist/mg;
    $line =~ s/<servicegroup>/$servicegroup/mg;
    print $CONFIG $line;
}

sub genHost {
    my $host = shift;
    my $CONFIG = shift;
    my $addNodetype = shift;
    my $addContact = shift;
    my $hostname = shift || $host->{hostname};
    my $address = shift || $host->{address};
    my $TEMPL;
    my $line;

    my $nodetypegroup = 'node-' . join ( ',node-', keys %{$host->{nodetypeArr}}) . ',site-' . $host->{sitename};
    $nodetypegroup .= ',' . $addNodetype if ($addNodetype);
    
    my $alias = $host->{sitename} . ' ' . $host->{nodetype};
    
    my $contactlist = $host->{contactlist};
    $contactlist .= ',' . $addContact if ($addContact);

    open ($TEMPL, "$TEMPLS/host.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<sitename>/$host->{sitename}/mg;
    $line =~ s/<alias>/$alias/mg;
    $line =~ s/<nodetype>/$host->{nodetype}/mg;
    $line =~ s/<nodetypegroup>/$nodetypegroup/mg;
    $line =~ s/<hostname>/$hostname/mg;
    $line =~ s/<address>/$address/mg;
    $line =~ s/<checktype>/$host->{checktype}/mg;
    $line =~ s/<contactlist>/$contactlist/mg;
    print $CONFIG $line;

}

sub genLocalHost {
    my $sitename = shift;
    my $host = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    my $address = inet_ntoa(scalar gethostbyname($host));

    open ($TEMPL, "$TEMPLS/host.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<sitename>/$sitename/mg;
    $line =~ s/<alias>/$sitename Nagios node/mg;
    $line =~ s/<nodetype>/Nagios/mg;
    $line =~ s/<nodetypegroup>/nagios-server/mg;
    $line =~ s/<hostname>/$host/mg;
    $line =~ s/<address>/$address/mg;
    $line =~ s/<checktype>/check-host-alive/mg;
    $line =~ s/<contactlist>/nagios-admins/mg;
    print $CONFIG $line;
    
    genCADist $CONFIG if ($PROBES ne 'remote');
}

sub genMyProxyHost {
    my $sitename = shift;
    my $host = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    my $address = inet_ntoa(scalar gethostbyname($host));

    open ($TEMPL, "$TEMPLS/host.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<sitename>/$sitename/mg;
    $line =~ s/<alias>/MyProxy server/mg;
    $line =~ s/<nodetype>/MyProxy/mg;
    $line =~ s/<nodetypegroup>/myproxy-server/mg;
    $line =~ s/<hostname>/$host/mg;
    $line =~ s/<address>/$address/mg;
    $line =~ s/<checktype>/check-tcp!7512/mg;
    $line =~ s/<contactlist>/nagios-admins/mg;
    print $CONFIG $line;
}

sub genServiceGroup {
    my $name = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;

    open ($TEMPL, "$TEMPLS/servicegroup.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<servicename>/$name/mg;
    
    print $CONFIG $line;
}

sub genServiceDependency {
    my $hostname1 = shift;
    my $servicename1 = shift;
    my $hostname2 = shift;
    my $servicename2 = shift;
    my $CONFIG = shift;
    my $TEMPL;
    my $line;
    
    open ($TEMPL, "$TEMPLS/servicedependency.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    $line =~ s/<servicename1>/$servicename1/mg;
    $line =~ s/<servicename2>/$servicename2/mg;
    $line =~ s/<hostname1>/$hostname1/mg;
    $line =~ s/<hostname2>/$hostname2/mg;

    print $CONFIG $line;
}


sub genService {
    my $probe = shift;
    my $service = shift;
    my $metric = shift;
    my $hostname = shift;
    my $contactlist = shift;
    my $serviceuri = shift;
    my $timeout = shift;
    my $interval = shift;
    my $retryInterval = shift;
    my $options = shift;
    my $line = shift;
    my $CONFIG = shift;

    $line =~ s/<hostname>/$hostname/mg;
    $line =~ s/<servicetype>/$service/mg;
    $line =~ s/<contactlist>/$contactlist/mg;
    $line =~ s/<probe>/$probe/mg;
    $line =~ s/<metric>/$metric/mg;
    $line =~ s/<serviceuri>/$serviceuri/mg;
    $line =~ s/<timeout>/$timeout/mg;
    $line =~ s/<options>/$options/mg;
    $line =~ s/<interval>/$interval/mg;
    $line =~ s/<retryInterval>/$retryInterval/mg;

    print $CONFIG $line;
}

sub genNativeService {
    my $probe = shift;
    my $service = shift;
    my $metric = shift;
    my $hostname = shift;
    my $contactlist = shift;
    my $timeout = shift;
    my $interval = shift;
    my $retryInterval = shift;
    my $options = shift;    
    my $line = shift;
    my $CONFIG = shift;

    $line =~ s/<hostname>/$hostname/mg;
    $line =~ s/<servicetype>/$service/mg;
    $line =~ s/<metric>/$metric/mg;
    $line =~ s/<contactlist>/$contactlist/mg;
    $line =~ s/<probe>/$probe/mg;
    $line =~ s/<options>/$options/mg;
    $line =~ s/<timeout>/$timeout/mg;
    $line =~ s/<interval>/$interval/mg;
    $line =~ s/<retryInterval>/$retryInterval/mg;

    print $CONFIG $line;
}

sub genServices {
    my $host = shift;
    my $local = shift;
    my $servicegroups = shift;
    my $CONFIG = shift;
    my $hostname = shift || $host->{hostname};    
    my $serviceuri;
    my $templWlcg;
    my $templNative;
    my $TEMPL;
    my $has_local = 0;

    open ($TEMPL, "$TEMPLS/service/wlcg.template");
    $templWlcg = join ("", <$TEMPL>);
    close ($TEMPL);

    open ($TEMPL, "$TEMPLS/service/native.template");
    $templNative = join ("", <$TEMPL>);
    close ($TEMPL);

    foreach my $service ( keys %{$host->{services}}){
        my $serviceuri = $host->{services}->{$service}->{serviceuri};
        $serviceuri =~ s/$host->{hostname}/$hostname/g if ($hostname ne $host->{hostname});
        foreach my $metric ( keys %{$WLCG_SERVICE->{$service}->{metrics}}) {
            my $currMetric = $WLCG_SERVICE->{$service}->{metrics}->{$metric};
            my $opts = "";
            
            # Create options string
            if (exists $currMetric->{attribute}->{TIMEOUT}) {
                $opts .= ' ' . $currMetric->{attribute}->{TIMEOUT} . ' ' . $currMetric->{config}->{timeout} . ' ';
            }
            foreach my $opt (keys %{$currMetric->{attribute}}) {
                if (exists $host->{$opt}) {
                    $opts .= ' ' . $currMetric->{attribute}->{$opt} . ' ' . $host->{$opt} . ' ';
                }
                if ($opt eq 'VONAME') {
                    $opts .= ' ' . $currMetric->{attribute}->{$opt} . ' ' . $VONAME . ' ';
                }
            }            
            
            if ($currMetric->{native}) {
                genNativeService  $currMetric->{probe}, $service, $metric, $hostname, $host->{contactlist},
                                  $currMetric->{config}->{timeout}, $currMetric->{config}->{interval},
                                  $currMetric->{config}->{retryInterval}, $opts, $templNative, $CONFIG;
            } else {
                genService  $currMetric->{probe}, $service, $metric, $hostname, $host->{contactlist},
                            $serviceuri, $currMetric->{config}->{timeout},
                            $currMetric->{config}->{interval}, $currMetric->{config}->{retryInterval}, $opts,
                            $templWlcg, $CONFIG;
            }
                        
            unless (exists $servicegroups->{$service}) {
                genServiceGroup $service, $CONFIG;
                $servicegroups->{$service} = 1;
            }

            foreach my $dep (@{$currMetric->{dependency}}) {
                if ($dep eq "hr.srce.GridProxy-Valid") {
                    genServiceDependency $hostname, $metric, $local, $dep, $CONFIG;
                } else {
                    genServiceDependency $hostname, $metric, $hostname, $dep, $CONFIG;
                }
            }
        }
	$has_local++;
}
    genHostAggregate $hostname, $host->{contactlist}, 'local', $CONFIG if $has_local;
    

}


sub genRemoteServices {
    my $host = shift;
    my $CONFIG = shift;
    my $hostname = shift || $host->{hostname};
    my $contactlist = shift || $host->{contactlist};
    my $TEMPL;
    my $line;

    my $has_sam = 0;
    my $has_npm = 0;

    open ($TEMPL, "$TEMPLS/service/remote.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    foreach my $servicename (keys %{$host->{remoteservices}}) {
        my $line2 = $line;
        $line2 =~ s/<hostname>/$hostname/mg;
        $line2 =~ s/<servicename>/$servicename/mg;
        $line2 =~ s/<contactlist>/$contactlist/mg;
        $line2 =~ s/<url>/$host->{remoteservices}->{$servicename}->{url}/mg;

        print $CONFIG $line2;
	$has_sam++;
    }
    
    genHostAggregate $hostname, $contactlist, 'sam', $CONFIG if $has_sam;
    
    open ($TEMPL, "$TEMPLS/service/npm.template");
    $line = join ("", <$TEMPL>);
    close ($TEMPL);

    foreach my $servicename (keys %{$host->{npmservices}}) {
        my $line2 = $line;
        $line2 =~ s/<hostname>/$hostname/mg;
        $line2 =~ s/<servicename>/$servicename/mg;
        $line2 =~ s/<contactlist>/$contactlist/mg;
	$line2 =~ s/<url>/$host->{npmservices}->{$servicename}->{url}/mg;

        print $CONFIG $line2;
	$has_npm++;
    }

    genHostAggregate $hostname, $contactlist, 'npm', $CONFIG if $has_npm;
}

sub genConfig {
    my $CONFIG;
    my $TEMPL;

    my %nodetypes;
    my $servicegroups = {};
    my $remoteServiceGroup = 0;
    my $localServiceGroup = 0;
    my $npmServiceGroup = 0;
    my $samServiceGroup = 0;
    my $aggregateServiceGroup = 0;

    verbose("Generating configuration:\n");

    my $localhost = hostname();
    my $username = getlogin();
    my $time = time2str("%Y-%m-%dT%H:%M:%SZ",time(),"UTC");
    open ($TEMPL, "$TEMPLS/master.template");
    open ($CONFIG, ">$OUTPUTFILE");
    my $line = join ("", <$TEMPL>);
    $line =~ s/<adminemail>/$ADMINEMAIL/g;
    $line =~ s/<generated_by>/$username/g;
    $line =~ s/<generated_time>/$time/g;
    $line =~ s/<command_line>/$COMMAND_LINE/g;
    print $CONFIG $line;
    close ($TEMPL);

    if ($PROBES ne 'remote') {
        genGridProxy $VONAME, $MYPROXY, $MYPROXY_NAME, $CONFIG;
        genMyProxy $MYPROXY_NAME, $CONFIG;
    }

    if ($PROBES ne 'local') {
        genRemote $GLOBAL_SITE, $CONFIG;
    }

    my $site = $GLOBAL_SITES->{$GLOBAL_SITE};

    #foreach my $site (values %$GLOBAL_SITES){
        genContact $site, $CONFIG;
        genHostGroupSiteNode $site->{sitename}, $CONFIG;

        foreach my $host (values %{$site->{hosts}}) {
            my $hostAddNodeType = "";
            my $hostAddContact = "";

            if ($host->{hostname} eq $localhost) {
                $hostAddNodeType = "nagios-server";
                $hostAddContact = "nagios-admins";
            }

            if ($PROBES ne 'remote' && $MYPROXY && $host->{hostname} eq $MYPROXY) {
                $hostAddContact = "nagios-admins";
                $hostAddNodeType .= "," if ($hostAddNodeType);
                $hostAddNodeType .= "myproxy-server";
            }

            genHost $host, $CONFIG, $hostAddNodeType, $hostAddContact;
	    genServices $host, $localhost, $servicegroups, $CONFIG if ($PROBES ne 'remote');
	    genRemoteServices $host, $CONFIG if ($PROBES ne 'local');

            # generate configuration for aliases
            if (exists $host->{aliases}) {
                my $hostAddNodeTypeTmp = $hostAddNodeType || '';
                $hostAddNodeTypeTmp .= "," if ($hostAddNodeTypeTmp);
                $hostAddNodeTypeTmp .= genHostGroupAliases($host->{hostname}, $CONFIG);
                foreach my $alias (keys %{$host->{aliases}}) {
                    genHost $host, $CONFIG, $hostAddNodeTypeTmp, $hostAddContact, $alias, $host->{aliases}->{$alias}->{address};
                    genServices $host, $localhost, $servicegroups, $CONFIG, $alias;
		    genRemoteServices $host->{aliases}->{$alias}, $CONFIG, $alias, $host->{contactlist} if ($PROBES ne 'local');
                }
            }

            # generate configuration for load-balanced nodes
            if (exists $host->{lbnodes}) {
                my $hostAddNodeTypeTmp = $hostAddNodeType || '';
                $hostAddNodeTypeTmp .= "," if ($hostAddNodeTypeTmp);
                $hostAddNodeTypeTmp .= genHostGroupLBNodes($host->{hostname}, $CONFIG);
                foreach my $lbnode (keys %{$host->{lbnodes}}) {
                    genHost $host, $CONFIG, $hostAddNodeTypeTmp, $hostAddContact, $lbnode, $host->{lbnodes}->{$lbnode}->{address};
                    genServices $host, $localhost, $servicegroups, $CONFIG, $lbnode;
                }
            }

            foreach my $nodeT (keys %{$host->{nodetypeArr}}) {
                unless (exists $nodetypes{$nodeT}) {
                    genHostGroupNodeType $nodeT, $CONFIG;
                    $nodetypes{$nodeT} = 1;
                }
            }
            
            # generate remote service group if remote services are present
            if (!$samServiceGroup && $host->{remoteservices} && ref($host->{remoteservices}) eq 'HASH') {
                genServiceGroup 'sam', $CONFIG;
                $samServiceGroup = 1;
                if (!$remoteServiceGroup) {
                    genServiceGroup 'remote', $CONFIG;
                    $remoteServiceGroup = 1;
                }
		if (!$aggregateServiceGroup) {
		    genServiceGroup 'aggregate', $CONFIG;
		    $aggregateServiceGroup = 1;
		}
            }
            
            # generate remote service group if remote services are present
            if (!$npmServiceGroup && $host->{npmservices} && ref($host->{npmservices}) eq 'HASH' ) {
                genServiceGroup 'npm', $CONFIG;
                $npmServiceGroup = 1;
                if (!$remoteServiceGroup) {
                    genServiceGroup 'remote', $CONFIG;
                    $remoteServiceGroup = 1;
                }
		if (!$aggregateServiceGroup) {
		    genServiceGroup 'aggregate', $CONFIG;
		    $aggregateServiceGroup = 1;
		}
            }
            
            # generate local service group if local services are present
            if (!$localServiceGroup && $host->{services} && %{$host->{services}}) {
                genServiceGroup 'local', $CONFIG;
                $localServiceGroup = 1;
		if (!$aggregateServiceGroup) {
		    genServiceGroup 'aggregate', $CONFIG;
		    $aggregateServiceGroup = 1;
		}
            }

            verbose("  Host: ".$host->{hostname}."\n");
            verbose("    nodetypes: ".$host->{nodetype}."\n");
            verbose("    services: ".join (", ", keys %{$host->{services}})."\n") if ($host->{services});
            verbose("    aliases: ".join (", ", keys %{$host->{aliases}})."\n") if ($host->{aliases});
            verbose("    load-balancing nodes: ".join (", ", keys %{$host->{lbnodes}})."\n") if ($host->{lbnodes});
        }

        if (! exists $site->{hosts}->{$localhost} ) {
            genLocalHost $GLOBAL_SITE, $localhost, $CONFIG;
        } else {
            if ( $PROBES ne 'remote' && ! exists $site->{hosts}->{$localhost}->{services}->{CAdistribution}) {
                genCADist $CONFIG;
            }
        }

        if ( $PROBES ne 'remote' && $MYPROXY && ! exists $site->{hosts}->{$MYPROXY} ) {
            genMyProxyHost $GLOBAL_SITE, $MYPROXY,$CONFIG;
        }
    #}

    close ($CONFIG);
    
    verbose("Done!\n");
}

###############################################################
#####
#####   Main program
#####
###############################################################

checkArgs;

getInfoSAM($GLOBAL_SITE);
$GLOBAL_SITES->{$GLOBAL_SITE}->{bdii} = $BDII if ($BDII && !$GLOBAL_SITES->{$GLOBAL_SITE}->{bdii});
if ($GLOBAL_SITES->{$GLOBAL_SITE}->{bdii}) {
    getInfoBDII($GLOBAL_SITE, $GLOBAL_SITES->{$GLOBAL_SITE}->{bdii}, $VONAME);
} else {
    verbose("Warning: Site BDII not found. Some local probes may not run correctly. Define BDII manually with option --bdii bdii.address.\n");
}

if ($PROBES ne 'remote') {
    getServices;
} else {
    getRemoteServicesOnly;
}

genConfig;




